package ca.recrown.islandsurvivalcraft.world.generation;

import org.bukkit.World;
import org.bukkit.block.Biome;

import ca.recrown.islandsurvivalcraft.Utilities;
import ca.recrown.islandsurvivalcraft.types.Point2;
import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateTargetValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.DepthFirstSearch;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;

public class UniBiomeIslandGenerator implements BiomeGenerator {
    private final Cache<Point2, Biome> biomeCache = new Cache<>(12800);
    private final Cache<Point2, Boolean> existenceCache = new Cache<>(12800);

    @Override
    public void generateBiomeColumn(Biome[][] biomes, World world, int chunkX, int chunkZ, int localX, int localZ, IslandWorldMapper mapper, BiomeSelector biomeSelector, TemperatureMapGenerator tempGen) {
        int worldX = 16 * chunkX + localX;
        int worldZ = 16 * chunkZ + localZ;
        Point2 chunkCoords = Utilities.worldToChunkCoordinates(new Point2(worldX, worldZ));
        existenceCache.setValue(chunkCoords, false);

        //Check if we can just give it something in cache.
        Biome biome = getSavedBiome(world, worldX, worldZ);
        if (biome != null) {
            biomes[localX][localZ] = biome;
            return;
        }

        //Fine, check if it's ocean.
        if (!mapper.isIsland(worldX, worldZ)) {
            setCacheBiome(worldX, worldZ, biomeSelector.getOceanBiome(tempGen.getTemperature(worldX, worldZ)), biomes);
            return;
        }

        //Shoot, looks like it's actually an island.
        DepthFirstSearch search = new DepthFirstSearch();
        search.setValidatable(mapper);
        
        IslandInfo islandInfo = new IslandInfo(mapper, world);
        search.setStartPosition(worldX, worldZ);
        if (!search.findTarget(islandInfo)) {
            float temp = tempGen.getTemperature(worldX, worldZ);
            if (islandInfo.main == null) islandInfo.main = biomeSelector.getLandBiome(temp);
            if (islandInfo.shore == null) islandInfo.shore = biomeSelector.getShoreBiome(islandInfo.main, temp);
            if (islandInfo.shallow == null) islandInfo.shallow = biomeSelector.getOceanBiome(temp);
        }

        PropagatorInfo propInfo = new PropagatorInfo(islandInfo, biomes, new Point2(chunkX, chunkZ), mapper);
        search.setValidatable(propInfo);
        search.setStartPosition(worldX, worldZ);
        search.findTarget(propInfo);
    }

    private Biome getSavedBiome(World world, int worldX, int worldZ) {
        Point2 worldCoords = new Point2(worldX, worldZ);
        Biome res = null;


        res = biomeCache.getValue(worldCoords);
        if (res != null) return res;

        Point2 chunkCoords = Utilities.worldToChunkCoordinates(worldCoords);
        Boolean chunkExists = existenceCache.getValue(chunkCoords);
        if (chunkExists == null) {
            chunkExists = world.isChunkGenerated(worldX, worldZ);
            existenceCache.setValue(chunkCoords, chunkExists);
        }

        if (chunkExists) {
            res = world.getBiome(worldX, 0, worldZ);
            biomeCache.setValue(worldCoords, res);
        }
        return res;
    }

    public void chunkGenerated(int chunkX, int chunkZ) {
        existenceCache.setValue(new Point2(chunkX, chunkZ), true);
    }

    private void setCacheBiome(int worldX, int worldZ, Biome biome, Biome[][] biomes) {
        Point2 worldCoords = new Point2(worldX, worldZ);
        if (biomes != null) {
            Point2 localCoords = Utilities.worldToLocalChunkCoordinates(worldCoords);
            biomes[localCoords.x][localCoords.y] = biome;
        }
        biomeCache.setValue(worldCoords, biome);
    }

    private class IslandInfo implements CoordinateTargetValidatable {
        public final IslandWorldMapper mapper;
        public final World world;
        public Biome main, shore, shallow;
        
        public IslandInfo(IslandWorldMapper mapper, World world) {
            this.mapper = mapper;
            this.world = world;
        }

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            if (mapper.isLand(x, y)) {
                if (mapper.isShore(x, y)) {
                    shore = getSavedBiome(world, x, y);
                } else {
                    main = getSavedBiome(world, x, y);
                }
            } else {
                shallow = getSavedBiome(world, x, y);
            }
            return main != null && shore != null && shallow != null;
        }
    }

    private class PropagatorInfo implements CoordinateTargetValidatable, CoordinateValidatable {
        private final Biome shallow, shore, main;
        private final Biome[][] biomes;
        private final Point2 chunkCoords;
        private final IslandWorldMapper mapper;

        public PropagatorInfo(IslandInfo islandInfo, Biome[][] biomes, Point2 chunkCoords, IslandWorldMapper mapper) {
            this.shallow = islandInfo.shallow;
            this.shore = islandInfo.shore;
            this.main = islandInfo.main;
            this.biomes = biomes;
            this.chunkCoords = chunkCoords;
            this.mapper = mapper;
        }

        @Override
        public boolean validate(int x, int y) {
            Point2 chunkCoords = Utilities.worldToChunkCoordinates(new Point2(x, y));
            return this.chunkCoords.fastEquals(chunkCoords) && mapper.validate(x, y);
        }

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            if (mapper.isLand(x, y)) {
                if (mapper.isShore(x, y)) {
                    setCacheBiome(x, y, shore, biomes);
                } else {
                    setCacheBiome(x, y, main, biomes);
                }
            } else {
                setCacheBiome(x, y, shallow, biomes);
            }
            return false;
        }

    }
}